import json

from bbot.logger import log_to_stderr
from bbot.modules.output.base import BaseOutputModule


class Stdout(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Output to text"}
    options = {"format": "text", "event_types": [], "event_fields": [], "in_scope_only": False, "accept_dupes": True}
    options_desc = {
        "format": "Which text format to display, choices: text,json",
        "event_types": "Which events to display, default all event types",
        "event_fields": "Which event fields to display",
        "in_scope_only": "Whether to only show in-scope events",
        "accept_dupes": "Whether to show duplicate events, default True",
    }
    vuln_severity_map = {"LOW": "HUGEWARNING", "MEDIUM": "HUGEWARNING", "HIGH": "CRITICAL", "CRITICAL": "CRITICAL"}
    format_choices = ["text", "json"]

    async def setup(self):
        self.text_format = self.config.get("format", "text").strip().lower()
        if not self.text_format in self.format_choices:
            return (
                False,
                f'Invalid text format choice, "{self.text_format}" (choices: {",".join(self.format_choices)})',
            )
        self.accept_event_types = [str(s).upper() for s in self.config.get("event_types", [])]
        self.show_event_fields = [str(s) for s in self.config.get("event_fields", [])]
        self.in_scope_only = self.config.get("in_scope_only", False)
        self.accept_dupes = self.config.get("accept_dupes", False)
        return True

    async def filter_event(self, event):
        if self.accept_event_types:
            if not event.type in self.accept_event_types:
                return False, f'Event type "{event.type}" is not in the allowed event_types'
        return True

    async def handle_event(self, event):
        json_mode = "human" if self.text_format == "text" else "json"
        event_json = event.json(mode=json_mode)
        if self.show_event_fields:
            event_json = {k: str(event_json.get(k, "")) for k in self.show_event_fields}

        if self.text_format == "text":
            await self.handle_text(event, event_json)
        elif self.text_format == "json":
            await self.handle_json(event, event_json)

    async def handle_text(self, event, event_json):
        if self.show_event_fields:
            event_str = "\t".join([str(s) for s in event_json.values()])
        else:
            event_str = self.human_event_str(event)

        # log vulnerabilities in vivid colors
        if event.type == "VULNERABILITY":
            severity = event.data.get("severity", "INFO")
            if severity in self.vuln_severity_map:
                loglevel = self.vuln_severity_map[severity]
                log_to_stderr(event_str, level=loglevel, logname=False)
        elif event.type == "FINDING":
            log_to_stderr(event_str, level="HUGEINFO", logname=False)

        print(event_str)

    async def handle_json(self, event, event_json):
        print(json.dumps(event_json))



from .base import ModuleTestBase


class TestPostman(ModuleTestBase):
    modules_overrides = ["postman", "speculate"]

    async def setup_after_prep(self, module_test):
        await module_test.mock_dns(
            {"blacklanternsecurity.com": {"A": ["127.0.0.99"]}, "github.com": {"A": ["127.0.0.99"]}}
        )
        module_test.httpx_mock.add_response(
            url="https://www.postman.com/_api/ws/proxy",
            json={
                "data": [
                    {
                        "score": 611.41156,
                        "normalizedScore": 23,
                        "document": {
                            "watcherCount": 6,
                            "apiCount": 0,
                            "forkCount": 0,
                            "isblacklisted": "false",
                            "createdAt": "2021-06-15T14:03:51",
                            "publishertype": "team",
                            "publisherHandle": "blacklanternsecurity",
                            "id": "11498add-357d-4bc5-a008-0a2d44fb8829",
                            "slug": "bbot-public",
                            "updatedAt": "2024-07-30T11:00:35",
                            "entityType": "workspace",
                            "visibilityStatus": "public",
                            "forkcount": "0",
                            "tags": [],
                            "createdat": "2021-06-15T14:03:51",
                            "forkLabel": "",
                            "publisherName": "blacklanternsecurity",
                            "name": "BlackLanternSecurity BBOT [Public]",
                            "dependencyCount": 7,
                            "collectionCount": 6,
                            "warehouse__updated_at": "2024-07-30 11:00:00",
                            "privateNetworkFolders": [],
                            "isPublisherVerified": False,
                            "publisherType": "team",
                            "curatedInList": [],
                            "creatorId": "6900157",
                            "description": "",
                            "forklabel": "",
                            "publisherId": "299401",
                            "publisherLogo": "",
                            "popularity": 5,
                            "isPublic": True,
                            "categories": [],
                            "universaltags": "",
                            "views": 5788,
                            "summary": "BLS public workspaces.",
                            "memberCount": 2,
                            "isBlacklisted": False,
                            "publisherid": "299401",
                            "isPrivateNetworkEntity": False,
                            "isDomainNonTrivial": True,
                            "privateNetworkMeta": "",
                            "updatedat": "2021-10-20T16:19:29",
                            "documentType": "workspace",
                        },
                        "highlight": {"summary": "<b>BLS</b> BBOT api test."},
                    }
                ],
                "meta": {
                    "queryText": "blacklanternsecurity",
                    "total": {
                        "collection": 0,
                        "request": 0,
                        "workspace": 1,
                        "api": 0,
                        "team": 0,
                        "user": 0,
                        "flow": 0,
                        "apiDefinition": 0,
                        "privateNetworkFolder": 0,
                    },
                    "state": "AQ4",
                    "spellCorrection": {"count": {"all": 1, "workspace": 1}, "correctedQueryText": None},
                    "featureFlags": {
                        "enabledPublicResultCuration": True,
                        "boostByPopularity": True,
                        "reRankPostNormalization": True,
                        "enableUrlBarHostNameSearch": True,
                    },
                },
            },
        )

    def check(self, module_test, events):
        assert len(events) == 4
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "blacklanternsecurity.com" and e.scope_distance == 0
            ]
        ), "Failed to emit target DNS_NAME"
        assert 1 == len(
            [e for e in events if e.type == "ORG_STUB" and e.data == "blacklanternsecurity" and e.scope_distance == 0]
        ), "Failed to find ORG_STUB"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "CODE_REPOSITORY"
                and "postman" in e.tags
                and e.data["url"] == "https://www.postman.com/blacklanternsecurity/bbot-public"
                and e.scope_distance == 1
            ]
        ), "Failed to find blacklanternsecurity postman workspace"



from .base import ModuleTestBase


class TestWayback(ModuleTestBase):
    async def setup_after_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url=f"http://web.archive.org/cdx/search/cdx?url=blacklanternsecurity.com&matchType=domain&output=json&fl=original&collapse=original",
            json=[["original"], ["http://asdf.blacklanternsecurity.com"]],
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



import zipfile
import json
from pathlib import Path
from bbot.modules.templates.postman import postman


class postman_download(postman):
    watched_events = ["CODE_REPOSITORY"]
    produced_events = ["FILESYSTEM"]
    flags = ["passive", "subdomain-enum", "safe", "code-enum"]
    meta = {
        "description": "Download workspaces, collections, requests from Postman",
        "created_date": "2024-09-07",
        "author": "@domwhewell-sage",
    }
    options = {"output_folder": "", "api_key": ""}
    options_desc = {"output_folder": "Folder to download postman workspaces to", "api_key": "Postman API Key"}
    scope_distance_modifier = 2

    async def setup(self):
        self.api_key = self.config.get("api_key", "")
        self.authorization_headers = {"X-Api-Key": self.api_key}

        output_folder = self.config.get("output_folder")
        if output_folder:
            self.output_dir = Path(output_folder) / "postman_workspaces"
        else:
            self.output_dir = self.scan.home / "postman_workspaces"
        self.helpers.mkdir(self.output_dir)
        return await self.require_api_key()

    async def ping(self):
        url = f"{self.api_url}/me"
        response = await self.helpers.request(url, headers=self.authorization_headers)
        assert getattr(response, "status_code", 0) == 200, response.text

    async def filter_event(self, event):
        if event.type == "CODE_REPOSITORY":
            if "postman" not in event.tags:
                return False, "event is not a postman workspace"
        return True

    async def handle_event(self, event):
        repo_url = event.data.get("url")
        workspace_id = await self.get_workspace_id(repo_url)
        if workspace_id:
            self.verbose(f"Found workspace ID {workspace_id} for {repo_url}")
            workspace_path = await self.download_workspace(workspace_id)
            if workspace_path:
                self.verbose(f"Downloaded workspace from {repo_url} to {workspace_path}")
                codebase_event = self.make_event(
                    {"path": str(workspace_path)}, "FILESYSTEM", tags=["postman", "workspace"], parent=event
                )
                await self.emit_event(
                    codebase_event,
                    context=f"{{module}} downloaded postman workspace at {repo_url} to {{event.type}}: {workspace_path}",
                )

    async def get_workspace_id(self, repo_url):
        workspace_id = ""
        profile = repo_url.split("/")[-2]
        name = repo_url.split("/")[-1]
        url = f"{self.base_url}/ws/proxy"
        json = {
            "service": "workspaces",
            "method": "GET",
            "path": f"/workspaces?handle={profile}&slug={name}",
        }
        r = await self.helpers.request(url, method="POST", json=json, headers=self.headers)
        if r is None:
            return workspace_id
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return workspace_id
        data = json.get("data", [])
        if len(data) == 1:
            workspace_id = data[0]["id"]
        return workspace_id

    async def download_workspace(self, id):
        zip_path = None
        workspace = await self.get_workspace(id)
        if workspace:
            # Create a folder for the workspace
            name = workspace["name"]
            folder = self.output_dir / name
            self.helpers.mkdir(folder)
            zip_path = folder / f"{id}.zip"

            # Main Workspace
            self.add_json_to_zip(zip_path, workspace, f"{name}.postman_workspace.json")

            # Workspace global variables
            self.verbose(f"Downloading globals for workspace {name}")
            globals = await self.get_globals(id)
            globals_id = globals["id"]
            self.add_json_to_zip(zip_path, globals, f"{globals_id}.postman_environment.json")

            # Workspace Environments
            workspace_environments = workspace.get("environments", [])
            if workspace_environments:
                self.verbose(f"Downloading environments for workspace {name}")
                for _ in workspace_environments:
                    environment_id = _["uid"]
                    environment = await self.get_environment(environment_id)
                    self.add_json_to_zip(zip_path, environment, f"{environment_id}.postman_environment.json")

            # Workspace Collections
            workspace_collections = workspace.get("collections", [])
            if workspace_collections:
                self.verbose(f"Downloading collections for workspace {name}")
                for _ in workspace_collections:
                    collection_id = _["uid"]
                    collection = await self.get_collection(collection_id)
                    self.add_json_to_zip(zip_path, collection, f"{collection_id}.postman_collection.json")
        return zip_path

    async def get_workspace(self, workspace_id):
        workspace = {}
        workspace_url = f"{self.api_url}/workspaces/{workspace_id}"
        r = await self.helpers.request(workspace_url, headers=self.authorization_headers)
        if r is None:
            return workspace
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return workspace
        workspace = json.get("workspace", {})
        return workspace

    async def get_globals(self, workspace_id):
        globals = {}
        globals_url = f"{self.base_url}/workspace/{workspace_id}/globals"
        r = await self.helpers.request(globals_url, headers=self.headers)
        if r is None:
            return globals
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return globals
        globals = json.get("data", {})
        return globals

    async def get_environment(self, environment_id):
        environment = {}
        environment_url = f"{self.api_url}/environments/{environment_id}"
        r = await self.helpers.request(environment_url, headers=self.authorization_headers)
        if r is None:
            return environment
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return environment
        environment = json.get("environment", {})
        return environment

    async def get_collection(self, collection_id):
        collection = {}
        collection_url = f"{self.api_url}/collections/{collection_id}"
        r = await self.helpers.request(collection_url, headers=self.authorization_headers)
        if r is None:
            return collection
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return collection
        collection = json.get("collection", {})
        return collection

    def add_json_to_zip(self, zip_path, data, filename):
        with zipfile.ZipFile(zip_path, "a") as zipf:
            json_content = json.dumps(data, indent=4)
            zipf.writestr(filename, json_content)



from .base import ModuleTestBase


class TestPostman_Download(ModuleTestBase):
    config_overrides = {"modules": {"postman_download": {"api_key": "asdf"}}}
    modules_overrides = ["postman", "postman_download", "speculate"]

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.getpostman.com/me",
            json={
                "user": {
                    "id": 000000,
                    "username": "test_key",
                    "email": "blacklanternsecurity@test.com",
                    "fullName": "Test Key",
                    "avatar": "",
                    "isPublic": True,
                    "teamId": 0,
                    "teamDomain": "",
                    "roles": ["user"],
                },
                "operations": [
                    {"name": "api_object_usage", "limit": 3, "usage": 0, "overage": 0},
                    {"name": "collection_run_limit", "limit": 25, "usage": 0, "overage": 0},
                    {"name": "file_storage_limit", "limit": 20, "usage": 0, "overage": 0},
                    {"name": "flow_count", "limit": 5, "usage": 0, "overage": 0},
                    {"name": "flow_requests", "limit": 5000, "usage": 0, "overage": 0},
                    {"name": "performance_test_limit", "limit": 25, "usage": 0, "overage": 0},
                    {"name": "postbot_calls", "limit": 50, "usage": 0, "overage": 0},
                    {"name": "reusable_packages", "limit": 3, "usage": 0, "overage": 0},
                    {"name": "test_data_retrieval", "limit": 1000, "usage": 0, "overage": 0},
                    {"name": "test_data_storage", "limit": 10, "usage": 0, "overage": 0},
                    {"name": "mock_usage", "limit": 1000, "usage": 0, "overage": 0},
                    {"name": "monitor_request_runs", "limit": 1000, "usage": 0, "overage": 0},
                    {"name": "api_usage", "limit": 1000, "usage": 0, "overage": 0},
                ],
            },
        )

    async def setup_after_prep(self, module_test):
        await module_test.mock_dns(
            {"blacklanternsecurity.com": {"A": ["127.0.0.99"]}, "github.com": {"A": ["127.0.0.99"]}}
        )
        module_test.httpx_mock.add_response(
            url="https://www.postman.com/_api/ws/proxy",
            match_content=b'{"service": "search", "method": "POST", "path": "/search-all", "body": {"queryIndices": ["collaboration.workspace"], "queryText": "blacklanternsecurity", "size": 100, "from": 0, "clientTraceId": "", "requestOrigin": "srp", "mergeEntities": "true", "nonNestedRequests": "true", "domain": "public"}}',
            json={
                "data": [
                    {
                        "score": 611.41156,
                        "normalizedScore": 23,
                        "document": {
                            "watcherCount": 6,
                            "apiCount": 0,
                            "forkCount": 0,
                            "isblacklisted": "false",
                            "createdAt": "2021-06-15T14:03:51",
                            "publishertype": "team",
                            "publisherHandle": "blacklanternsecurity",
                            "id": "11498add-357d-4bc5-a008-0a2d44fb8829",
                            "slug": "bbot-public",
                            "updatedAt": "2024-07-30T11:00:35",
                            "entityType": "workspace",
                            "visibilityStatus": "public",
                            "forkcount": "0",
                            "tags": [],
                            "createdat": "2021-06-15T14:03:51",
                            "forkLabel": "",
                            "publisherName": "blacklanternsecurity",
                            "name": "BlackLanternSecurity BBOT [Public]",
                            "dependencyCount": 7,
                            "collectionCount": 6,
                            "warehouse__updated_at": "2024-07-30 11:00:00",
                            "privateNetworkFolders": [],
                            "isPublisherVerified": False,
                            "publisherType": "team",
                            "curatedInList": [],
                            "creatorId": "6900157",
                            "description": "",
                            "forklabel": "",
                            "publisherId": "299401",
                            "publisherLogo": "",
                            "popularity": 5,
                            "isPublic": True,
                            "categories": [],
                            "universaltags": "",
                            "views": 5788,
                            "summary": "BLS public workspaces.",
                            "memberCount": 2,
                            "isBlacklisted": False,
                            "publisherid": "299401",
                            "isPrivateNetworkEntity": False,
                            "isDomainNonTrivial": True,
                            "privateNetworkMeta": "",
                            "updatedat": "2021-10-20T16:19:29",
                            "documentType": "workspace",
                        },
                        "highlight": {"summary": "<b>BLS</b> BBOT api test."},
                    }
                ],
                "meta": {
                    "queryText": "blacklanternsecurity",
                    "total": {
                        "collection": 0,
                        "request": 0,
                        "workspace": 1,
                        "api": 0,
                        "team": 0,
                        "user": 0,
                        "flow": 0,
                        "apiDefinition": 0,
                        "privateNetworkFolder": 0,
                    },
                    "state": "AQ4",
                    "spellCorrection": {"count": {"all": 1, "workspace": 1}, "correctedQueryText": None},
                    "featureFlags": {
                        "enabledPublicResultCuration": True,
                        "boostByPopularity": True,
                        "reRankPostNormalization": True,
                        "enableUrlBarHostNameSearch": True,
                    },
                },
            },
        )
        module_test.httpx_mock.add_response(
            url="https://www.postman.com/_api/ws/proxy",
            match_content=b'{"service": "workspaces", "method": "GET", "path": "/workspaces?handle=blacklanternsecurity&slug=bbot-public"}',
            json={
                "meta": {"model": "workspace", "action": "find", "nextCursor": ""},
                "data": [
                    {
                        "id": "3a7e4bdc-7ff7-4dd4-8eaa-61ddce1c3d1b",
                        "name": "BlackLanternSecurity BBOT [Public]",
                        "description": None,
                        "summary": "BLS public workspaces.",
                        "createdBy": "299401",
                        "updatedBy": "299401",
                        "team": None,
                        "createdAt": "2021-10-20T16:19:29",
                        "updatedAt": "2021-10-20T16:19:29",
                        "visibilityStatus": "public",
                        "profileInfo": {
                            "slug": "bbot-public",
                            "profileType": "team",
                            "profileId": "000000",
                            "publicHandle": "https://www.postman.com/blacklanternsecurity",
                            "publicImageURL": "",
                            "publicName": "BlackLanternSecurity",
                            "isVerified": False,
                        },
                    }
                ],
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.getpostman.com/workspaces/3a7e4bdc-7ff7-4dd4-8eaa-61ddce1c3d1b",
            json={
                "workspace": {
                    "id": "3a7e4bdc-7ff7-4dd4-8eaa-61ddce1c3d1b",
                    "name": "BlackLanternSecurity BBOT [Public]",
                    "type": "personal",
                    "description": None,
                    "visibility": "public",
                    "createdBy": "00000000",
                    "updatedBy": "00000000",
                    "createdAt": "2021-11-17T06:09:01.000Z",
                    "updatedAt": "2021-11-17T08:57:16.000Z",
                    "collections": [
                        {
                            "id": "2aab9fd0-3715-4abe-8bb0-8cb0264d023f",
                            "name": "BBOT Public",
                            "uid": "10197090-2aab9fd0-3715-4abe-8bb0-8cb0264d023f",
                        },
                    ],
                    "environments": [
                        {
                            "id": "f770f816-9c6a-40f7-bde3-c0855d2a1089",
                            "name": "BBOT Test",
                            "uid": "10197090-f770f816-9c6a-40f7-bde3-c0855d2a1089",
                        }
                    ],
                    "apis": [],
                }
            },
        )
        module_test.httpx_mock.add_response(
            url="https://www.postman.com/_api/workspace/3a7e4bdc-7ff7-4dd4-8eaa-61ddce1c3d1b/globals",
            json={
                "model_id": "8be7574b-219f-49e0-8d25-da447a882e4e",
                "meta": {"model": "globals", "action": "find"},
                "data": {
                    "workspace": "3a7e4bdc-7ff7-4dd4-8eaa-61ddce1c3d1b",
                    "lastUpdatedBy": "00000000",
                    "lastRevision": 1637239113000,
                    "id": "8be7574b-219f-49e0-8d25-da447a882e4e",
                    "values": [
                        {
                            "key": "endpoint_url",
                            "value": "https://api.blacklanternsecurity.com/",
                            "enabled": True,
                        },
                    ],
                    "createdAt": "2021-11-17T06:09:01.000Z",
                    "updatedAt": "2021-11-18T12:38:33.000Z",
                },
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.getpostman.com/environments/10197090-f770f816-9c6a-40f7-bde3-c0855d2a1089",
            json={
                "environment": {
                    "id": "f770f816-9c6a-40f7-bde3-c0855d2a1089",
                    "name": "BBOT Test",
                    "owner": "00000000",
                    "createdAt": "2021-11-17T06:29:54.000Z",
                    "updatedAt": "2021-11-23T07:06:53.000Z",
                    "values": [
                        {
                            "key": "temp_session_endpoint",
                            "value": "https://api.blacklanternsecurity.com/",
                            "enabled": True,
                        },
                    ],
                    "isPublic": True,
                }
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.getpostman.com/collections/10197090-62b91565-d2e2-4bcd-8248-4dba2e3452f0",
            json={
                "collection": {
                    "info": {
                        "_postman_id": "62b91565-d2e2-4bcd-8248-4dba2e3452f0",
                        "name": "BBOT Public",
                        "schema": "https://schema.getpostman.com/json/collection/v2.1.0/collection.json",
                        "updatedAt": "2021-11-17T07:13:16.000Z",
                        "createdAt": "2021-11-17T07:13:15.000Z",
                        "lastUpdatedBy": "00000000",
                        "uid": "00000000-62b91565-d2e2-4bcd-8248-4dba2e3452f0",
                    },
                    "item": [
                        {
                            "name": "Generate API Session",
                            "id": "c1bac38c-dfc9-4cc0-9c19-828cbc8543b1",
                            "protocolProfileBehavior": {"disableBodyPruning": True},
                            "request": {
                                "method": "POST",
                                "header": [{"key": "Content-Type", "value": "application/json"}],
                                "body": {
                                    "mode": "raw",
                                    "raw": '{"username": "test", "password": "Test"}',
                                },
                                "url": {"raw": "{{endpoint_url}}", "host": ["{{endpoint_url}}"]},
                                "description": "",
                            },
                            "response": [],
                            "uid": "10197090-c1bac38c-dfc9-4cc0-9c19-828cbc8543b1",
                        },
                    ],
                }
            },
        )

    def check(self, module_test, events):
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "CODE_REPOSITORY"
                and "postman" in e.tags
                and e.data["url"] == "https://www.postman.com/blacklanternsecurity/bbot-public"
                and e.scope_distance == 1
            ]
        ), "Failed to find blacklanternsecurity postman workspace"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "FILESYSTEM"
                and "postman_workspaces/BlackLanternSecurity BBOT [Public]" in e.data["path"]
                and "postman" in e.tags
                and e.scope_distance == 1
            ]
        ), "Failed to find blacklanternsecurity postman workspace"



import os
import time
import logging

from .misc import sha1

log = logging.getLogger("bbot.core.helpers.cache")


def cache_get(self, key, text=True, cache_hrs=24 * 7):
    """
    Get an item from the cache. Default expiration is 1 week.
    Returns None if item is not in cache
    """
    filename = self.cache_filename(key)
    if filename.is_file():
        valid = self.is_cached(key, cache_hrs)
        if valid:
            open_kwargs = {}
            if text:
                open_kwargs.update({"mode": "r", "encoding": "utf-8", "errors": "ignore"})
            else:
                open_kwargs["mode"] = "rb"
            log.debug(f'Using cached content for "{key}"')
            return open(filename, **open_kwargs).read()
        else:
            log.debug(f'Cached content for "{key}" is older than {cache_hrs:,} hours')


def cache_put(self, key, content):
    """
    Put an item in the cache.
    """
    filename = self.cache_filename(key)
    if type(content) == bytes:
        open_kwargs = {"mode": "wb"}
    else:
        open_kwargs = {"mode": "w", "encoding": "utf-8"}
        content = str(content)
    with open(filename, **open_kwargs) as f:
        f.write(content)


def is_cached(self, key, cache_hrs=24 * 7):
    filename = self.cache_filename(key)
    if filename.is_file():
        (m, i, d, n, u, g, sz, atime, mtime, ctime) = os.stat(filename)
        return mtime > time.time() - cache_hrs * 3600
    return False


def cache_filename(self, key):
    return self.cache_dir / sha1(key).hexdigest()



from bbot.modules.templates.bucket import bucket_template


class bucket_firebase(bucket_template):
    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum", "web-basic"]
    meta = {
        "description": "Check for open Firebase databases related to target",
        "created_date": "2023-03-20",
        "author": "@TheTechromancer",
    }
    options = {"permutations": False}
    options_desc = {
        "permutations": "Whether to try permutations",
    }

    cloud_helper_name = "google"
    delimiters = ("", "-")
    base_domains = ["firebaseio.com"]

    def filter_bucket(self, event):
        host = str(event.host)
        if not any(host.endswith(f".{d}") for d in self.base_domains):
            return False, "bucket belongs to a different cloud provider"
        return True, ""

    def build_url(self, bucket_name, base_domain, region):
        return f"https://{bucket_name}.{base_domain}/.json"

    async def check_bucket_open(self, bucket_name, url):
        url = url.strip("/") + "/.json"
        response = await self.helpers.request(url)
        tags = self.gen_tags_exists(response)
        status_code = getattr(response, "status_code", 404)
        msg = ""
        if status_code == 200:
            msg = "Open storage bucket"
        return (msg, tags)



from bbot.modules.base import BaseModule


class Ipstack(BaseModule):
    """
    Ipstack GeoIP
    Leverages the ipstack.com API to geolocate a host by IP address.
    """

    watched_events = ["IP_ADDRESS"]
    produced_events = ["GEOLOCATION"]
    flags = ["passive", "safe"]
    meta = {
        "description": "Query IPStack's GeoIP API",
        "created_date": "2022-11-26",
        "author": "@tycoonslive",
        "auth_required": True,
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "IPStack GeoIP API Key"}
    scope_distance_modifier = 1
    _priority = 2
    suppress_dupes = False

    base_url = "http://api.ipstack.com"

    async def setup(self):
        return await self.require_api_key()

    async def ping(self):
        url = f"{self.base_url}/check?access_key={self.api_key}"
        r = await self.request_with_fail_count(url)
        resp_content = getattr(r, "text", "")
        assert getattr(r, "status_code", 0) == 200, resp_content

    async def handle_event(self, event):
        try:
            url = f"{self.base_url}/{event.data}?access_key={self.api_key}"
            result = await self.request_with_fail_count(url)
            if result:
                geo_data = result.json()
                if not geo_data:
                    self.verbose(f"No JSON response from {url}")
            else:
                self.verbose(f"No response from {url}")
        except Exception:
            self.verbose(f"Error retrieving results for {event.data}", trace=True)
            return
        geo_data = {k: v for k, v in geo_data.items() if v is not None}
        if "error" in geo_data:
            error_msg = geo_data.get("error").get("info", "")
            if error_msg:
                self.warning(error_msg)
        elif geo_data:
            country = geo_data.get("country_name", "unknown country")
            region = geo_data.get("region_name", "unknown region")
            city = geo_data.get("city", "unknown city")
            lat = geo_data.get("latitude", "")
            long = geo_data.get("longitude", "")
            description = f"{city}, {region}, {country} ({lat}, {long})"
            await self.emit_event(
                geo_data,
                "GEOLOCATION",
                event,
                context=f'{{module}} queried ipstack.com\'s API for "{event.data}" and found {{event.type}}: {description}',
            )



from .base import ModuleTestBase


class TestGit(ModuleTestBase):
    targets = [
        "http://127.0.0.1:8888/",
        "http://127.0.0.1:8888/test/asdf",
        "http://127.0.0.1:8888/test2",
    ]

    modules_overrides = ["git", "httpx"]

    git_config = """[core]
    repositoryformatversion = 0
    filemode = true
    bare = false
    logallrefupdates = true"""

    async def setup_after_prep(self, module_test):
        module_test.set_expect_requests(
            expect_args={"uri": "/.git/config"}, respond_args={"response_data": self.git_config}
        )
        module_test.set_expect_requests(
            expect_args={"uri": "/test/.git/config"}, respond_args={"response_data": self.git_config}
        )
        module_test.set_expect_requests(
            expect_args={"uri": "/test/asdf/.git/config"}, respond_args={"response_data": self.git_config}
        )
        module_test.set_expect_requests(expect_args={"uri": "/test2/.git/config"}, respond_args={"response_data": ""})

    def check(self, module_test, events):
        assert any(
            e.type == "FINDING" and "http://127.0.0.1:8888/.git/config" in e.data["description"] for e in events
        )
        assert any(
            e.type == "FINDING" and "http://127.0.0.1:8888/test/.git/config" in e.data["description"] for e in events
        )
        assert any(
            e.type == "FINDING" and "http://127.0.0.1:8888/test/asdf/.git/config" in e.data["description"]
            for e in events
        )
        assert not any(
            e.type == "FINDING" and "http://127.0.0.1:8888/test2/.git/config" in e.data["description"] for e in events
        )



from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class virustotal(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query VirusTotal's API for subdomains",
        "created_date": "2022-08-25",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "VirusTotal API Key"}

    base_url = "https://www.virustotal.com/api/v3"

    async def setup(self):
        self.api_key = self.config.get("api_key", "")
        self.headers = {"x-apikey": self.api_key}
        return await super().setup()

    async def ping(self):
        # virustotal does not have a ping function
        return

    def parse_results(self, r, query):
        results = set()
        text = getattr(r, "text", "")
        for match in self.helpers.regexes.dns_name_regex.findall(text):
            match = match.lower()
            if match.endswith(query):
                results.add(match)
        return results

    async def query(self, query):
        results = set()
        url = f"{self.base_url}/domains/{self.helpers.quote(query)}/subdomains"
        agen = self.helpers.api_page_iter(
            url, json=False, headers=self.headers, next_key=lambda r: r.json().get("links", {}).get("next", "")
        )
        try:
            async for response in agen:
                r = self.parse_results(response, query)
                if not r:
                    break
                results.update(r)
        finally:
            agen.aclose()
        return results



from bbot.modules.base import BaseModule


class gitlab(BaseModule):
    watched_events = ["HTTP_RESPONSE", "TECHNOLOGY", "SOCIAL"]
    produced_events = ["TECHNOLOGY", "SOCIAL", "CODE_REPOSITORY", "FINDING"]
    flags = ["active", "safe", "code-enum"]
    meta = {
        "description": "Detect GitLab instances and query them for repositories",
        "created_date": "2024-03-11",
        "author": "@TheTechromancer",
    }
    options = {"api_key": ""}
    options_desc = {"api_key": "Gitlab access token"}

    scope_distance_modifier = 2

    async def setup(self):
        self.headers = {}
        self.api_key = self.config.get("api_key", "")
        if self.api_key:
            self.headers.update({"Authorization": f"Bearer {self.api_key}"})
        return True

    async def filter_event(self, event):
        # only accept out-of-scope SOCIAL events
        if event.type == "HTTP_RESPONSE":
            if event.scope_distance > self.scan.scope_search_distance:
                return False, "event is out of scope distance"
        elif event.type == "TECHNOLOGY":
            if not event.data["technology"].lower().startswith("gitlab"):
                return False, "technology is not gitlab"
            if not self.helpers.is_ip(event.host) and self.helpers.tldextract(event.host).domain == "gitlab":
                return False, "gitlab instance is not self-hosted"
        elif event.type == "SOCIAL":
            if event.data["platform"] != "gitlab":
                return False, "platform is not gitlab"
        return True

    async def handle_event(self, event):
        if event.type == "HTTP_RESPONSE":
            await self.handle_http_response(event)
        elif event.type == "TECHNOLOGY":
            await self.handle_technology(event)
        elif event.type == "SOCIAL":
            await self.handle_social(event)

    async def handle_http_response(self, event):
        # identify gitlab instances from HTTP responses
        # HTTP_RESPONSE --> TECHNOLOGY
        # HTTP_RESPONSE --> FINDING
        headers = event.data.get("header", {})
        if "x_gitlab_meta" in headers:
            url = event.parsed_url._replace(path="/").geturl()
            await self.emit_event(
                {"host": str(event.host), "technology": "GitLab", "url": url},
                "TECHNOLOGY",
                parent=event,
                context=f"{{module}} detected {{event.type}}: GitLab at {url}",
            )
            description = f"GitLab server at {event.host}"
            await self.emit_event(
                {"host": str(event.host), "description": description},
                "FINDING",
                parent=event,
                context=f"{{module}} detected {{event.type}}: {description}",
            )

    async def handle_technology(self, event):
        # retrieve gitlab groups from gitlab instances
        # TECHNOLOGY --> SOCIAL
        # TECHNOLOGY --> URL
        # TECHNOLOGY --> CODE_REPOSITORY
        base_url = self.get_base_url(event)
        projects_url = self.helpers.urljoin(base_url, "api/v4/projects?simple=true")
        await self.handle_projects_url(projects_url, event)
        groups_url = self.helpers.urljoin(base_url, "api/v4/groups?simple=true")
        await self.handle_groups_url(groups_url, event)

    async def handle_social(self, event):
        # retrieve repositories from gitlab user
        # SOCIAL --> CODE_REPOSITORY
        # SOCIAL --> SOCIAL
        username = event.data.get("profile_name", "")
        if not username:
            return
        base_url = self.get_base_url(event)
        urls = [
            # group
            self.helpers.urljoin(base_url, f"api/v4/users/{username}/projects?simple=true"),
            # user
            self.helpers.urljoin(base_url, f"api/v4/groups/{username}/projects?simple=true"),
        ]
        for url in urls:
            await self.handle_projects_url(url, event)

    async def handle_projects_url(self, projects_url, event):
        for project in await self.gitlab_json_request(projects_url):
            project_url = project.get("web_url", "")
            if project_url:
                code_event = self.make_event({"url": project_url}, "CODE_REPOSITORY", tags="git", parent=event)
                await self.emit_event(
                    code_event, context=f"{{module}} enumerated projects and found {{event.type}} at {project_url}"
                )
            namespace = project.get("namespace", {})
            if namespace:
                await self.handle_namespace(namespace, event)

    async def handle_groups_url(self, groups_url, event):
        for group in await self.gitlab_json_request(groups_url):
            await self.handle_namespace(group, event)

    async def gitlab_json_request(self, url):
        response = await self.helpers.request(url, headers=self.headers)
        if response is not None:
            try:
                json = response.json()
            except Exception:
                return []
            if json and isinstance(json, list):
                return json
        return []

    async def handle_namespace(self, namespace, event):
        namespace_name = namespace.get("path", "")
        namespace_url = namespace.get("web_url", "")
        namespace_path = namespace.get("full_path", "")
        if namespace_name and namespace_url and namespace_path:
            namespace_url = self.helpers.parse_url(namespace_url)._replace(path=f"/{namespace_path}").geturl()
            social_event = self.make_event(
                {"platform": "gitlab", "profile_name": namespace_path, "url": namespace_url},
                "SOCIAL",
                parent=event,
            )
            await self.emit_event(
                social_event,
                context=f'{{module}} found GitLab namespace ({{event.type}}) "{namespace_name}" at {namespace_url}',
            )

    def get_base_url(self, event):
        base_url = event.data.get("url", "")
        if not base_url:
            base_url = f"https://{event.host}"
        return self.helpers.urlparse(base_url)._replace(path="/").geturl()



import os
import dns
import time
import asyncio
import logging
import traceback
from cachetools import LRUCache
from contextlib import suppress

from bbot.core.engine import EngineServer
from bbot.core.helpers.async_helpers import NamedLock
from bbot.core.helpers.dns.helpers import extract_targets
from bbot.core.helpers.misc import (
    is_ip,
    rand_string,
    parent_domain,
    domain_parents,
)


log = logging.getLogger("bbot.core.helpers.dns.engine.server")

all_rdtypes = ["A", "AAAA", "SRV", "MX", "NS", "SOA", "CNAME", "TXT"]


class DNSEngine(EngineServer):

    CMDS = {
        0: "resolve",
        1: "resolve_raw",
        2: "resolve_batch",
        3: "resolve_raw_batch",
        4: "is_wildcard",
        5: "is_wildcard_domain",
        99: "_mock_dns",
    }

    def __init__(self, socket_path, config={}, debug=False):
        super().__init__(socket_path, debug=debug)

        self.config = config
        self.dns_config = self.config.get("dns", {})
        # config values
        self.timeout = self.dns_config.get("timeout", 5)
        self.retries = self.dns_config.get("retries", 1)
        self.abort_threshold = self.dns_config.get("abort_threshold", 50)

        # resolver
        self.resolver = dns.asyncresolver.Resolver()
        self.resolver.rotate = True
        self.resolver.timeout = self.timeout
        self.resolver.lifetime = self.timeout

        # skip certain queries
        dns_omit_queries = self.dns_config.get("omit_queries", None)
        if not dns_omit_queries:
            dns_omit_queries = []
        self.dns_omit_queries = dict()
        for d in dns_omit_queries:
            d = d.split(":")
            if len(d) == 2:
                rdtype, query = d
                rdtype = rdtype.upper()
                query = query.lower()
                try:
                    self.dns_omit_queries[rdtype].add(query)
                except KeyError:
                    self.dns_omit_queries[rdtype] = {query}

        # wildcard handling
        self.wildcard_ignore = self.dns_config.get("wildcard_ignore", None)
        if not self.wildcard_ignore:
            self.wildcard_ignore = []
        self.wildcard_ignore = tuple([str(d).strip().lower() for d in self.wildcard_ignore])
        self.wildcard_tests = self.dns_config.get("wildcard_tests", 5)
        self._wildcard_cache = dict()
        # since wildcard detection takes some time, This is to prevent multiple
        # modules from kicking off wildcard detection for the same domain at the same time
        self._wildcard_lock = NamedLock()

        self._dns_connectivity_lock = None
        self._last_dns_success = None
        self._last_connectivity_warning = time.time()
        # keeps track of warnings issued for wildcard detection to prevent duplicate warnings
        self._dns_warnings = set()
        self._errors = dict()
        self._debug = self.dns_config.get("debug", False)
        self._dns_cache = LRUCache(maxsize=10000)

        self.filter_bad_ptrs = self.dns_config.get("filter_ptrs", True)

    async def resolve(self, query, **kwargs):
        """Resolve DNS names and IP addresses to their corresponding results.

        This is a high-level function that can translate a given domain name to its associated IP addresses
        or an IP address to its corresponding domain names. It's structured for ease of use within modules
        and will abstract away most of the complexity of DNS resolution, returning a simple set of results.

        Args:
            query (str): The domain name or IP address to resolve.
            **kwargs: Additional arguments to be passed to the resolution process.

        Returns:
            set: A set containing resolved domain names or IP addresses.

        Examples:
            >>> results = await resolve("1.2.3.4")
            {"evilcorp.com"}

            >>> results = await resolve("evilcorp.com")
            {"1.2.3.4", "dead::beef"}
        """
        results = set()
        try:
            answers, errors = await self.resolve_raw(query, **kwargs)
            for answer in answers:
                for _, host in extract_targets(answer):
                    results.add(host)
        except BaseException:
            self.log.trace(f"Caught exception in resolve({query}, {kwargs}):")
            self.log.trace(traceback.format_exc())
            raise

        self.debug(f"Results for {query} with kwargs={kwargs}: {results}")
        return results

    async def resolve_raw(self, query, **kwargs):
        """Resolves the given query to its associated DNS records.

        This function is a foundational method for DNS resolution in this class. It understands both IP addresses and
        hostnames and returns their associated records in a raw format provided by the dnspython library.

        Args:
            query (str): The IP address or hostname to resolve.
            type (str or list[str], optional): Specifies the DNS record type(s) to fetch. Can be a single type like 'A'
                or a list like ['A', 'AAAA']. If set to 'any', 'all', or '*', it fetches all supported types. If not
                specified, the function defaults to fetching 'A' and 'AAAA' records.
            **kwargs: Additional arguments that might be passed to the resolver.

        Returns:
            tuple: A tuple containing two lists:
                - list: A list of tuples where each tuple consists of a record type string (like 'A') and the associated
                  raw dnspython answer.
                - list: A list of tuples where each tuple consists of a record type string and the associated error if
                  there was an issue fetching the record.

        Examples:
            >>> await resolve_raw("8.8.8.8")
            ([('PTR', <dns.resolver.Answer object at 0x7f4a47cdb1d0>)], [])

            >>> await resolve_raw("dns.google")
            (<dns.resolver.Answer object at 0x7f4a47ce46d0>, [])
        """
        # DNS over TCP is more reliable
        # But setting this breaks DNS resolution on Ubuntu because systemd-resolve doesn't support TCP
        # kwargs["tcp"] = True
        try:
            query = str(query).strip()
            kwargs.pop("rdtype", None)
            rdtype = kwargs.pop("type", "A")
            if is_ip(query):
                return await self._resolve_ip(query, **kwargs)
            else:
                return await self._resolve_hostname(query, rdtype=rdtype, **kwargs)
        except BaseException:
            self.log.trace(f"Caught exception in resolve_raw({query}, {kwargs}):")
            self.log.trace(traceback.format_exc())
            raise

    async def _resolve_hostname(self, query, **kwargs):
        """Translate a hostname into its corresponding IP addresses.

        This is the foundational function for converting a domain name into its associated IP addresses. It's designed
        for internal use within the class and handles retries, caching, and a variety of error/timeout scenarios.
        It also respects certain configurations that might ask to skip certain types of queries. Results are returned
        in the default dnspython answer object format.

        Args:
            query (str): The hostname to resolve.
            rdtype (str, optional): The type of DNS record to query (e.g., 'A', 'AAAA'). Defaults to 'A'.
            retries (int, optional): The number of times to retry on failure. Defaults to class-wide `retries`.
            use_cache (bool, optional): Whether to check the cache before trying a fresh resolution. Defaults to True.
            **kwargs: Additional arguments that might be passed to the resolver.

        Returns:
            tuple: A tuple containing:
                - list: A list of resolved IP addresses.
                - list: A list of errors encountered during the resolution process.

        Examples:
            >>> results, errors = await _resolve_hostname("google.com")
            (<dns.resolver.Answer object at 0x7f4a4b2caf50>, [])
        """
        self.debug(f"Resolving {query} with kwargs={kwargs}")
        results = []
        errors = []
        rdtype = kwargs.get("rdtype", "A")

        # skip certain queries if requested
        if rdtype in self.dns_omit_queries:
            if any(h == query or query.endswith(f".{h}") for h in self.dns_omit_queries[rdtype]):
                self.debug(f"Skipping {rdtype}:{query} because it's omitted in the config")
                return results, errors

        parent = parent_domain(query)
        retries = kwargs.pop("retries", self.retries)
        use_cache = kwargs.pop("use_cache", True)
        tries_left = int(retries) + 1
        parent_hash = hash((parent, rdtype))
        dns_cache_hash = hash((query, rdtype))
        while tries_left > 0:
            try:
                if use_cache:
                    results = self._dns_cache.get(dns_cache_hash, [])
                if not results:
                    error_count = self._errors.get(parent_hash, 0)
                    if error_count >= self.abort_threshold:
                        connectivity = await self._connectivity_check()
                        if connectivity:
                            self.log.verbose(
                                f'Aborting query "{query}" because failed {rdtype} queries for "{parent}" ({error_count:,}) exceeded abort threshold ({self.abort_threshold:,})'
                            )
                            if parent_hash not in self._dns_warnings:
                                self.log.verbose(
                                    f'Aborting future {rdtype} queries to "{parent}" because error count ({error_count:,}) exceeded abort threshold ({self.abort_threshold:,})'
                                )
                            self._dns_warnings.add(parent_hash)
                            return results, errors
                    results = await self._catch(self.resolver.resolve, query, **kwargs)
                    if use_cache:
                        self._dns_cache[dns_cache_hash] = results
                    if parent_hash in self._errors:
                        self._errors[parent_hash] = 0
                break
            except (
                dns.resolver.NoNameservers,
                dns.exception.Timeout,
                dns.resolver.LifetimeTimeout,
                TimeoutError,
                asyncio.exceptions.TimeoutError,
            ) as e:
                try:
                    self._errors[parent_hash] += 1
                except KeyError:
                    self._errors[parent_hash] = 1
                errors.append(e)
                # don't retry if we get a SERVFAIL
                if isinstance(e, dns.resolver.NoNameservers):
                    break
                tries_left -= 1
                err_msg = (
                    f'DNS error or timeout for {rdtype} query "{query}" ({self._errors[parent_hash]:,} so far): {e}'
                )
                if tries_left > 0:
                    retry_num = (retries + 1) - tries_left
                    self.debug(err_msg)
                    self.debug(f"Retry (#{retry_num}) resolving {query} with kwargs={kwargs}")
                else:
                    self.log.verbose(err_msg)

        if results:
            self._last_dns_success = time.time()
            self.debug(f"Answers for {query} with kwargs={kwargs}: {list(results)}")

        if errors:
            self.debug(f"Errors for {query} with kwargs={kwargs}: {errors}")

        return results, errors

    async def _resolve_ip(self, query, **kwargs):
        """Translate an IP address into a corresponding DNS name.

        This is the most basic function that will convert an IP address into its associated domain name. It handles
        retries, caching, and multiple types of timeout/error scenarios internally. The function is intended for
        internal use and should not be directly called by modules without understanding its intricacies.

        Args:
            query (str): The IP address to be reverse-resolved.
            retries (int, optional): The number of times to retry on failure. Defaults to 0.
            use_cache (bool, optional): Whether to check the cache for the result before attempting resolution. Defaults to True.
            **kwargs: Additional arguments to be passed to the resolution process.

        Returns:
            tuple: A tuple containing:
                - list: A list of resolved domain names (in default dnspython answer format).
                - list: A list of errors encountered during resolution.

        Examples:
            >>> results, errors = await _resolve_ip("8.8.8.8")
            (<dns.resolver.Answer object at 0x7f4a47cdb1d0>, [])
        """
        self.debug(f"Reverse-resolving {query} with kwargs={kwargs}")
        retries = kwargs.pop("retries", 0)
        use_cache = kwargs.pop("use_cache", True)
        tries_left = int(retries) + 1
        results = []
        errors = []
        dns_cache_hash = hash((query, "PTR"))
        while tries_left > 0:
            try:
                if use_cache:
                    results = self._dns_cache.get(dns_cache_hash, [])
                if not results:
                    results = await self._catch(self.resolver.resolve_address, query, **kwargs)
                    if use_cache:
                        self._dns_cache[dns_cache_hash] = results
                break
            except (
                dns.resolver.NoNameservers,
                dns.exception.Timeout,
                dns.resolver.LifetimeTimeout,
                TimeoutError,
                asyncio.exceptions.TimeoutError,
            ) as e:
                errors.append(e)
                # don't retry if we get a SERVFAIL
                if isinstance(e, dns.resolver.NoNameservers):
                    self.debug(f"{e} (query={query}, kwargs={kwargs})")
                    break
                else:
                    tries_left -= 1
                    if tries_left > 0:
                        retry_num = (retries + 2) - tries_left
                        self.debug(f"Retrying (#{retry_num}) {query} with kwargs={kwargs}")

        if results:
            self._last_dns_success = time.time()

        return results, errors

    async def resolve_batch(self, queries, threads=10, **kwargs):
        """
        A helper to execute a bunch of DNS requests.

        Args:
            queries (list): List of queries to resolve.
            **kwargs: Additional keyword arguments to pass to `resolve()`.

        Yields:
            tuple: A tuple containing the original query and its resolved value.

        Examples:
            >>> import asyncio
            >>> async def example_usage():
            ...     async for result in resolve_batch(['www.evilcorp.com', 'evilcorp.com']):
            ...         print(result)
            ('www.evilcorp.com', {'1.1.1.1'})
            ('evilcorp.com', {'2.2.2.2'})
        """
        async for (args, _, _), responses in self.task_pool(
            self.resolve, args_kwargs=queries, threads=threads, global_kwargs=kwargs
        ):
            yield args[0], responses

    async def resolve_raw_batch(self, queries, threads=10, **kwargs):
        queries_kwargs = [[q[0], {"type": q[1]}] for q in queries]
        async for (args, kwargs, _), (answers, errors) in self.task_pool(
            self.resolve_raw, args_kwargs=queries_kwargs, threads=threads, global_kwargs=kwargs
        ):
            query = args[0]
            rdtype = kwargs["type"]
            yield ((query, rdtype), (answers, errors))

    async def _catch(self, callback, *args, **kwargs):
        """
        Asynchronously catches exceptions thrown during DNS resolution and logs them.

        This method wraps around a given asynchronous callback function to handle different
        types of DNS exceptions and general exceptions. It logs the exceptions for debugging
        and, in some cases, re-raises them.

        Args:
            callback (callable): The asynchronous function to be executed.
            *args: Positional arguments to pass to the callback.
            **kwargs: Keyword arguments to pass to the callback.

        Returns:
            Any: The return value of the callback function, or an empty list if an exception is caught.

        Raises:
            dns.resolver.NoNameservers: When no nameservers could be reached.
        """
        try:
            return await callback(*args, **kwargs)
        except dns.resolver.NoNameservers:
            raise
        except (dns.exception.Timeout, dns.resolver.LifetimeTimeout, TimeoutError):
            self.log.debug(f"DNS query with args={args}, kwargs={kwargs} timed out after {self.timeout} seconds")
            raise
        except dns.exception.DNSException as e:
            self.debug(f"{e} (args={args}, kwargs={kwargs})")
        except Exception as e:
            self.log.warning(f"Error in {callback.__qualname__}() with args={args}, kwargs={kwargs}: {e}")
            self.log.trace(traceback.format_exc())
        return []

    async def is_wildcard(self, query, rdtypes, raw_dns_records=None):
        """
        Use this method to check whether a *host* is a wildcard entry

        This can reliably tell the difference between a valid DNS record and a wildcard within a wildcard domain.

        It works by making a bunch of random DNS queries to the parent domain, compiling a list of wildcard IPs,
        then comparing those to the IPs of the host in question. If the host's IP matches the wildcard ones, it's a wildcard.

        If you want to know whether a domain is using wildcard DNS, use `is_wildcard_domain()` instead.

        Args:
            query (str): The hostname to check for a wildcard entry.
            rdtypes (list): The DNS record type (e.g., "A", "AAAA") to consider during the check.
            raw_dns_records (dict, optional): Dictionary of {rdtype: [answer1, answer2, ...], ...} containing raw dnspython answers for the query.

        Returns:
            dict: A dictionary indicating if the query is a wildcard for each checked DNS record type.
                Keys are DNS record types like "A", "AAAA", etc.
                Values are tuples where the first element is a boolean indicating if the query is a wildcard,
                and the second element is the wildcard parent if it's a wildcard.

        Examples:
            >>> is_wildcard("www.github.io", rdtypes=["A", "AAAA", "MX"])
            {"A": (True, "github.io"), "AAAA": (True, "github.io"), "MX": (False, "github.io")}

            >>> is_wildcard("www.evilcorp.com", rdtypes=["A"])
            {"A": (False, "evilcorp.com")}

        Note:
            `is_wildcard` can be True, False, or None (indicating that wildcard detection was inconclusive)
        """
        if isinstance(rdtypes, str):
            rdtypes = [rdtypes]

        result = {}

        # if the work of resolving hasn't been done yet, do it
        if raw_dns_records is None:
            raw_dns_records = {}
            queries = [(query, rdtype) for rdtype in rdtypes]
            async for (_, rdtype), (answers, errors) in self.resolve_raw_batch(queries):
                if answers:
                    for answer in answers:
                        try:
                            raw_dns_records[rdtype].add(answer)
                        except KeyError:
                            raw_dns_records[rdtype] = {answer}
                else:
                    if errors:
                        self.debug(f"Failed to resolve {query} ({rdtype}) during wildcard detection")
                        result[rdtype] = ("ERROR", query)

        # clean + process the raw records into a baseline
        baseline = {}
        baseline_raw = {}
        for rdtype, answers in raw_dns_records.items():
            for answer in answers:
                text_answer = answer.to_text()
                try:
                    baseline_raw[rdtype].add(text_answer)
                except KeyError:
                    baseline_raw[rdtype] = {text_answer}
                for _, host in extract_targets(answer):
                    try:
                        baseline[rdtype].add(host)
                    except KeyError:
                        baseline[rdtype] = {host}

        # if it's unresolved, it's a big nope
        if not raw_dns_records:
            return result

        # once we've resolved the base query and have IP addresses to work with
        # we can compare the IPs to the ones we have on file for wildcards

        # only bother to check the rdypes that actually resolve
        rdtypes_to_check = set(raw_dns_records)

        # for every parent domain, starting with the shortest
        parents = list(domain_parents(query))
        for parent in parents[::-1]:

            # check if the parent domain is set up with wildcards
            wildcard_results = await self.is_wildcard_domain(parent, rdtypes_to_check)

            # for every rdtype
            for rdtype in list(baseline_raw):
                # skip if we already found a wildcard for this rdtype
                if rdtype in result:
                    continue

                # get our baseline IPs from above
                _baseline = baseline.get(rdtype, set())
                _baseline_raw = baseline_raw.get(rdtype, set())

                wildcard_rdtypes = wildcard_results.get(parent, {})
                wildcards = wildcard_rdtypes.get(rdtype, None)
                if wildcards is None:
                    continue
                wildcards, wildcard_raw = wildcards

                if wildcard_raw:
                    # skip this rdtype from now on
                    rdtypes_to_check.remove(rdtype)

                    # check if any of our baseline IPs are in the wildcard results
                    is_wildcard = any(r in wildcards for r in _baseline)
                    is_wildcard_raw = any(r in wildcard_raw for r in _baseline_raw)

                    # if there are any matches, we have a wildcard
                    if is_wildcard or is_wildcard_raw:
                        result[rdtype] = (True, parent)
                    else:
                        # otherwise, it's still suspicious, because we had random stuff resolve at this level
                        result[rdtype] = ("POSSIBLE", parent)

        # any rdtype that wasn't a wildcard, mark it as False
        for rdtype, answers in baseline_raw.items():
            if answers and rdtype not in result:
                result[rdtype] = (False, query)

        return result

    async def is_wildcard_domain(self, domain, rdtypes):
        """
        Check whether a given host or its children make use of wildcard DNS entries. Wildcard DNS can have
        various implications, particularly in subdomain enumeration and subdomain takeovers.

        Args:
            domain (str): The domain to check for wildcard DNS entries.
            rdtypes (list): Which DNS record types to check.

        Returns:
            dict: A dictionary where the keys are the parent domains that have wildcard DNS entries,
            and the values are another dictionary of DNS record types ("A", "AAAA", etc.) mapped to
            sets of their resolved IP addresses.

        Examples:
            >>> is_wildcard_domain("github.io")
            {"github.io": {"A": {"1.2.3.4"}, "AAAA": {"dead::beef"}}}

            >>> is_wildcard_domain("example.com")
            {}
        """
        if isinstance(rdtypes, str):
            rdtypes = [rdtypes]
        rdtypes = set(rdtypes)

        wildcard_results = {}
        # make a list of its parents
        parents = list(domain_parents(domain, include_self=True))
        # and check each of them, beginning with the highest parent (i.e. the root domain)
        for i, host in enumerate(parents[::-1]):
            host_results = {}
            queries = [((host, rdtype), {}) for rdtype in rdtypes]
            async for ((_, rdtype), _, _), (results, results_raw) in self.task_pool(
                self._is_wildcard_zone, args_kwargs=queries
            ):
                # if we hit a wildcard, we can skip this rdtype from now on
                if results_raw:
                    rdtypes.remove(rdtype)
                    host_results[rdtype] = results, results_raw

            if host_results:
                wildcard_results[host] = host_results

        return wildcard_results

    async def _is_wildcard_zone(self, host, rdtype):
        """
        Check whether a specific DNS zone+rdtype has a wildcard configuration
        """
        rdtype = rdtype.upper()

        # have we checked this host before?
        host_hash = hash((host, rdtype))
        async with self._wildcard_lock.lock(host_hash):
            # if we've seen this host before
            try:
                wildcard_results, wildcard_results_raw = self._wildcard_cache[host_hash]
                self.debug(f"Got {host}:{rdtype} from cache")
            except KeyError:
                wildcard_results = set()
                wildcard_results_raw = set()
                self.debug(f"Checking if {host}:{rdtype} is a wildcard")

                # determine if this is a wildcard domain
                # resolve a bunch of random subdomains of the same parent
                rand_queries = []
                for _ in range(self.wildcard_tests):
                    rand_query = f"{rand_string(digits=False, length=10)}.{host}"
                    rand_queries.append((rand_query, rdtype))

                async for (query, rdtype), (answers, errors) in self.resolve_raw_batch(rand_queries, use_cache=False):
                    for answer in answers:
                        # consider both the raw record
                        wildcard_results_raw.add(answer.to_text())
                        # and all the extracted hosts
                        for _, t in extract_targets(answer):
                            wildcard_results.add(t)

                if wildcard_results:
                    self.log.info(f"Encountered domain with wildcard DNS ({rdtype}): *.{host}")
                else:
                    self.debug(f"Finished checking {host}:{rdtype}, it is not a wildcard")
                self._wildcard_cache[host_hash] = wildcard_results, wildcard_results_raw

        return wildcard_results, wildcard_results_raw

    async def _is_wildcard(self, query, rdtypes, dns_children):
        if isinstance(rdtypes, str):
            rdtypes = [rdtypes]

    @property
    def dns_connectivity_lock(self):
        if self._dns_connectivity_lock is None:
            self._dns_connectivity_lock = asyncio.Lock()
        return self._dns_connectivity_lock

    async def _connectivity_check(self, interval=5):
        """
        Periodically checks for an active internet connection by attempting DNS resolution.

        Args:
            interval (int, optional): The time interval, in seconds, at which to perform the check.
            Defaults to 5 seconds.

        Returns:
            bool: True if there is an active internet connection, False otherwise.

        Examples:
            >>> await _connectivity_check()
            True
        """
        if self._last_dns_success is not None:
            if time.time() - self._last_dns_success < interval:
                return True
        dns_server_working = []
        async with self.dns_connectivity_lock:
            with suppress(Exception):
                dns_server_working = await self._catch(self.resolver.resolve, "www.google.com", rdtype="A")
                if dns_server_working:
                    self._last_dns_success = time.time()
                    return True
        if time.time() - self._last_connectivity_warning > interval:
            self.log.warning(f"DNS queries are failing, please check your internet connection")
            self._last_connectivity_warning = time.time()
        self._errors.clear()
        return False

    def debug(self, *args, **kwargs):
        if self._debug:
            self.log.trace(*args, **kwargs)

    @property
    def in_tests(self):
        return os.getenv("BBOT_TESTING", "") == "True"

    async def _mock_dns(self, mock_data, custom_lookup_fn=None):
        from .mock import MockResolver

        def deserialize_function(func_source):
            assert self.in_tests, "Can only mock when BBOT_TESTING=True"
            if func_source is None:
                return None
            exec(func_source)
            return locals()["custom_lookup"]

        self.resolver = MockResolver(mock_data, custom_lookup_fn=deserialize_function(custom_lookup_fn))



from bbot.modules.base import BaseModule


class postman(BaseModule):
    """
    A template module for use of the GitHub API
    Inherited by several other github modules.
    """

    base_url = "https://www.postman.com/_api"
    api_url = "https://api.getpostman.com"
    html_url = "https://www.postman.com"

    headers = {
        "Content-Type": "application/json",
        "X-App-Version": "10.18.8-230926-0808",
        "X-Entity-Team-Id": "0",
        "Origin": "https://www.postman.com",
        "Referer": "https://www.postman.com/search?q=&scope=public&type=all",
    }
